"""
Each motor is associated with a :class:`~.Calibration` that maps arbitrary
real-world coordinates to devices coordinates. When a calibration is associated
with an motor, the position can be changed with :meth:`Motor.set_position` and
:meth:`Motor.move`::

    from concert.devices.motors.dummy import Motor

    motor = Motor()

    motor.position = 2 * q.mm
    motor.move(-0.5 * q.mm)

As long as an motor is moving, :meth:`Motor.stop` will stop the motion.
"""
import asyncio
import logging
from concert.coroutines.base import background
from concert.quantities import q
from concert.base import Quantity, State, check, AccessorNotImplementedError
from concert.devices.base import Device
from concert.config import MOTOR_VELOCITY_SAMPLING_TIME as dT


LOG = logging.getLogger(__name__)


class _PositionMixin(Device):

    """Provide positional, discrete behaviour interface."""

    def __init__(self):
        super(_PositionMixin, self).__init__()

    @background
    async def move(self, delta):
        """
        move(delta)

        Move motor by *delta* user units."""
        await self.set_position(await self.get_position() + delta)

    @background
    @check(source=['hard-limit', 'moving'], target='standby')
    async def stop(self):
        """
        stop()

        Stop the motion."""
        await self._stop()

    @background
    @check(source='*', target=['standby', 'hard-limit'])
    async def home(self):
        """
        home()

        Home motor.
        """
        await self._home()

    @background
    @check(source='disabled', target='standby')
    async def enable(self):
        await self._enable()

    @background
    @check(source='standby', target='disabled')
    async def disable(self):
        await self._disable()

    async def _enable(self):
        raise AccessorNotImplementedError

    async def _disable(self):
        raise AccessorNotImplementedError

    async def _home(self):
        raise AccessorNotImplementedError

    async def _stop(self):
        raise AccessorNotImplementedError

    async def _emergency_stop(self):
        """Abort by default stops. If the motor provides a real emergency stop, _abort should be
        overriden.
        """
        if await self.get_state() == 'moving':
            await self.stop()


class LinearMotor(_PositionMixin):

    """
    One-dimensional linear motor.

    .. attribute:: position

        Position of the motor in length units.
    """

    def __init__(self):
        super(LinearMotor, self).__init__()

    async def _get_state(self):
        raise NotImplementedError

    state = State(default='standby')

    position = Quantity(q.mm, help="Position",
                        check=check(source=['hard-limit', 'standby'],
                                    target=['hard-limit', 'standby']))


class _VelocityMixin(object):

    """
    Provides velocity calculation for continuous motors
    from subsequent position measures.

    Sampling time for velocity measurement can be set in
    concert.config.MOTOR_VELOCITY_SAMPLING_TIME.
    """

    async def _get_velocity(self):
        pos0 = await self.get_position()
        await asyncio.sleep(dT.to(q.s).magnitude)
        pos1 = await self.get_position()

        return (pos1 - pos0) / dT


class ContinuousLinearMotor(LinearMotor, _VelocityMixin):

    """
    One-dimensional linear motor with adjustable velocity.

    .. attribute:: velocity

        Current velocity in length per time unit.
    """

    def __init__(self):
        super(ContinuousLinearMotor, self).__init__()

    async def _get_state(self):
        raise NotImplementedError

    state = State(default='standby')

    velocity = Quantity(q.mm / q.s, help="Linear velocity",
                        check=check(source=['hard-limit', 'standby', 'moving'],
                                    target=['moving', 'standby']))


class RotationMotor(_PositionMixin):

    """
    One-dimensional rotational motor.

    .. attribute:: position

        Position of the motor in angular units.
    """

    async def _get_state(self):
        raise NotImplementedError

    state = State(default='standby')

    position = Quantity(q.deg, help="Angular position",
                        check=check(source=['hard-limit', 'standby'],
                                    target=['hard-limit', 'standby']))

    def __init__(self):
        super(RotationMotor, self).__init__()


class ContinuousRotationMotor(RotationMotor, _VelocityMixin):

    """
    One-dimensional rotational motor with adjustable velocity.

    .. attribute:: velocity

        Current velocity in angle per time unit.
    """

    def __init__(self):
        super(ContinuousRotationMotor, self).__init__()

    async def _get_state(self):
        raise NotImplementedError

    state = State(default='standby')

    velocity = Quantity(q.deg / q.s, help="Angular velocity",
                        check=check(source=['hard-limit', 'standby', 'moving'],
                                    target=['moving', 'standby']))



'''
Attenuator module.
Getter and Setter for filter material and the
current state must be provided by the device implementation.
'''

from concert.base import State, AccessorNotImplementedError
from concert.devices.base import Device


class AttenuatorBox(Device):
    '''Attenuator box base class'''
    state = State(default='standby')

    def __init__(self):
        super(AttenuatorBox, self).__init__()

    async def _set_attenuator(self, att):
        raise AccessorNotImplementedError

    async def _get_attenuator(self):
        raise AccessorNotImplementedError


class AttenuatorBoxError(Exception):
    ''' Attenuator related error'''
    pass



from concert.tests import TestCase
from concert.base import HardLimitError
from concert.quantities import q
from concert.devices.motors.dummy import LinearMotor


class TestIssue280(TestCase):

    def test_hard_limit_throws_off_fsm(self):
        m = LinearMotor(upper_hard_limit=3 * q.mm)

        with self.assertRaises(HardLimitError):
            m.position = 5 * q.m

        m.position = 2 * q.mm



from concert.base import TransitionNotAllowed
from concert.devices.grippers.dummy import Gripper
from concert.tests import TestCase


class TestGripper(TestCase):

    def setUp(self):
        self.gripper = Gripper()

    async def test_grip(self):
        if await self.gripper.get_state() != 'released':
            await self.gripper.release()

        await self.gripper.grip()
        self.assertEqual('gripped', await self.gripper.get_state())
        with self.assertRaises(TransitionNotAllowed):
            await self.gripper.grip()

    async def test_release(self):
        if await self.gripper.get_state() != 'gripped':
            await self.gripper.grip()

        await self.gripper.release()
        self.assertEqual('released', await self.gripper.get_state())
        with self.assertRaises(TransitionNotAllowed):
            await self.gripper.release()



'''
Monochromator module. The device implementation needs to provide getters and
setters for wither wavelength or energy, it does not matter which one. The
conversion is handled in the base class.
'''
from concert.quantities import q
from concert.base import Quantity, AccessorNotImplementedError, State, check
from concert.devices.base import Device

# pint supports constants and defines hbar like this, but I haven't found a way
# to get the constants ...
hbar = 6.62606957e-34 * q.J * q.s


def energy_to_wavelength(energy):
    """Convert *energy* [eV-like] to wavelength [m]."""
    res = hbar * q.c / energy
    res.ito(q.m)
    return res


def wavelength_to_energy(wavelength):
    """Convert wavelength [m-like] to energy [eV]."""
    res = hbar * q.c / wavelength
    res.ito(q.eV)
    return res


class Monochromator(Device):

    """Monochromator device which is used to filter the beam in order to
    get a very narrow energy bandwidth.

    .. py:attribute:: energy

        Monochromatic energy in electron volts.

    .. py:attribute:: wavelength

        Monochromatic wavelength in meters.
    """

    state = State(default="standby")
    energy = Quantity(q.eV, help="Energy")
    wavelength = Quantity(q.nanometer, help="Wavelength")

    async def _get_energy(self):
        try:
            return await self._get_energy_real()
        except AccessorNotImplementedError:
            return wavelength_to_energy(await self._get_wavelength_real())

    @check(source="standby", target="standby")
    async def _set_energy(self, energy):
        try:
            await self._set_energy_real(energy)
        except AccessorNotImplementedError:
            await self._set_wavelength_real(energy_to_wavelength(energy))

    async def _get_wavelength(self):
        try:
            return await self._get_wavelength_real()
        except AccessorNotImplementedError:
            return energy_to_wavelength(await self._get_energy_real())

    @check(source="standby", target="standby")
    async def _set_wavelength(self, wavelength):
        try:
            await self._set_wavelength_real(wavelength)
        except AccessorNotImplementedError:
            await self._set_energy_real(wavelength_to_energy(wavelength))

    async def _get_energy_real(self):
        raise AccessorNotImplementedError

    async def _set_energy_real(self, energy):
        raise AccessorNotImplementedError

    async def _get_wavelength_real(self):
        raise AccessorNotImplementedError

    async def _set_wavelength_real(self, wavelength):
        raise AccessorNotImplementedError



"""Motor Dummy."""
import asyncio
from concert.base import HardLimitError, Quantity
from concert.coroutines.base import start
from concert.quantities import q
from concert.devices.motors import base
import numpy as np

MOVEMENT_TIME_STEPS = 0.01 * q.s


class _PositionMixin(object):
    def __init__(self):
        self._position = 0 * q.mm
        self._moving = False
        self._stop_evt = asyncio.Event()
        self._lower_hard_limit = -np.inf * q.mm
        self._upper_hard_limit = np.inf * q.mm

    async def _set_position(self, position):
        self._stop_evt.clear()
        try:
            direction = 0
            motion_velocity = await self.get_motion_velocity()
            if self._position < position:
                direction = 1
            elif self._position > position:
                direction = -1

            if direction:
                self._moving = True
                while (direction * self._position) < (direction * position) and self._moving:
                    self._position += direction * motion_velocity * MOVEMENT_TIME_STEPS
                    await asyncio.sleep(MOVEMENT_TIME_STEPS.to(q.s).magnitude)
                    if self._position < self._lower_hard_limit:
                        self._position = self._lower_hard_limit
                        raise HardLimitError('hard-limit')
                    if self._position > self._upper_hard_limit:
                        self._position = self._upper_hard_limit
                        raise HardLimitError('hard-limit')
                if self._moving:
                    self._position = position
        except asyncio.CancelledError:
            # fall-through to finally
            pass
        finally:
            self._moving = False
            self._stop_evt.set()

    async def _get_position(self):
        return self._position

    async def _home(self):
        print('start homing')
        await asyncio.sleep(1)
        await self.set_position(0 * q.mm)
        print('homes')

    async def _stop(self):
        self._moving = False
        await self._stop_evt.wait()

    async def _set_motion_velocity(self, vel):
        self._motion_velocity = vel

    async def _get_motion_velocity(self):
        return self._motion_velocity

    async def _get_state(self):
        if self._moving:
            return 'moving'
        if self._position > self._lower_hard_limit and self._position < self._upper_hard_limit:
            return 'standby'
        return 'hard-limit'


class LinearMotor(_PositionMixin, base.LinearMotor):

    """A linear step motor dummy."""

    motion_velocity = Quantity(q.mm / q.s)

    def __init__(self, position=None, upper_hard_limit=None, lower_hard_limit=None):
        base.LinearMotor.__init__(self)
        _PositionMixin.__init__(self)
        self._motion_velocity = 200000 * q.mm / q.s

        if position:
            self._position = position
        if upper_hard_limit:
            self._upper_hard_limit = upper_hard_limit
        if lower_hard_limit:
            self._lower_hard_limit = lower_hard_limit


class ContinuousLinearMotor(LinearMotor, base.ContinuousLinearMotor):

    """A continuous linear motor dummy."""

    def __init__(self, position=None, upper_hard_limit=None, lower_hard_limit=None):
        base.ContinuousLinearMotor.__init__(self)
        LinearMotor.__init__(self)
        if position:
            self._position = position
        if upper_hard_limit:
            self._upper_hard_limit = upper_hard_limit
        if lower_hard_limit:
            self._lower_hard_limit = lower_hard_limit

    async def _set_velocity(self, vel):
        if vel.magnitude > 0:
            await self.set_motion_velocity(vel)
            if self['position'].upper:
                start(self.set_position(self['position'].upper))
            else:
                start(self.set_position(np.inf * q.mm))
        if vel.magnitude < 0:
            await self.set_motion_velocity(vel)
            if self['position'].lower:
                start(self.set_position(self['position'].lower))
            else:
                start(self.set_position(-np.inf * q.mm))
        if vel.magnitude == 0:
            start(self.stop())


class RotationMotor(_PositionMixin, base.RotationMotor):

    """A rotational step motor dummy."""

    motion_velocity = Quantity(q.deg / q.s)

    def __init__(self, upper_hard_limit=None, lower_hard_limit=None):
        base.RotationMotor.__init__(self)
        _PositionMixin.__init__(self)
        self._position = 0 * q.deg
        self._lower_hard_limit = -np.inf * q.deg
        self._upper_hard_limit = np.inf * q.deg
        self._motion_velocity = 50000 * q.deg / q.s
        if upper_hard_limit:
            self._upper_hard_limit = upper_hard_limit
        if lower_hard_limit:
            self._lower_hard_limit = lower_hard_limit


class ContinuousRotationMotor(RotationMotor,
                              base.ContinuousRotationMotor):

    """A continuous rotational step motor dummy."""

    def __init__(self, position=None, upper_hard_limit=None, lower_hard_limit=None):
        base.ContinuousRotationMotor.__init__(self)
        RotationMotor.__init__(self)
        if position:
            self._position = position
        if upper_hard_limit:
            self._upper_hard_limit = upper_hard_limit
        if lower_hard_limit:
            self._lower_hard_limit = lower_hard_limit

    async def _set_velocity(self, vel):
        if vel.magnitude > 0:
            await self.set_motion_velocity(vel)
            if self['position'].upper:
                start(self.set_position(self['position'].upper))
            else:
                start(self.set_position(np.inf * q.deg))
        if vel.magnitude < 0:
            await self.set_motion_velocity(vel)
            if self['position'].lower:
                start(self.set_position(self['position'].lower))
            else:
                start(self.set_position(-np.inf * q.deg))
        if vel.magnitude == 0:
            start(self.stop())



"""A dummy X-ray tube."""

from concert.base import transition
from concert.devices.xraytubes import base
from concert.quantities import q


class XRayTube(base.XRayTube):

    """A dummy X-ray tube implementation."""

    def __init__(self):
        super(XRayTube, self).__init__()
        self._voltage = 0 * q.V
        self._current = 0 * q.A

    @transition(target='on')
    async def _on(self):
        pass

    @transition(target='off')
    async def _off(self):
        pass

    async def _get_voltage(self):
        return self._voltage

    async def _set_voltage(self, voltage):
        self._voltage = voltage

    async def _get_current(self):
        return self._current

    async def _set_current(self, current):
        self._current = current



from concert.quantities import q
from concert import optimization
from concert.tests import slow, assert_almost_equal, TestCase
from concert.optimization import optimize_parameter
from concert.devices.motors.dummy import LinearMotor


class TestOptimizers(TestCase):

    def setUp(self):
        super(TestOptimizers, self).setUp()
        self.algorithms = [optimization.halver, optimization.down_hill,
                           optimization.powell,
                           optimization.nonlinear_conjugate,
                           optimization.bfgs,
                           optimization.least_squares]
        self.center = 1.0 * q.mm
        self.motor = LinearMotor(position=0 * q.mm)
        self.motor.lower = -float('Inf') * q.mm
        self.motor.upper = float('Inf') * q.mm

    async def feedback(self):
        return ((await self.motor.get_position()).to(q.mm).magnitude
                - self.center.to(q.mm).magnitude) ** 2

    async def check(self):
        assert_almost_equal((await self.motor.get_position()).to(q.mm),
                            self.center.to(q.mm), 1e-2)

    async def optimize(self, algorithm):
        await optimize_parameter(self.motor["position"], self.feedback,
                                 await self.motor.get_position(), algorithm)

    @slow
    async def test_algorithms(self):
        for i in range(len(self.algorithms)):
            await self.motor.set_position(0 * q.mm)
            await self.optimize(self.algorithms[i])
            await self.check()



from concert.quantities import q
from concert.devices.motors.dummy import LinearMotor, ContinuousLinearMotor
from concert.devices.motors.dummy import RotationMotor, ContinuousRotationMotor
from concert.tests import TestCase, assert_almost_equal


class TestMotor(TestCase):

    def setUp(self):
        super(TestMotor, self).setUp()
        self.motor = LinearMotor()

    def test_set_position(self):
        position = 1 * q.mm
        self.motor.position = position
        self.assertEqual(position, self.motor.position)

    async def test_move(self):
        position = 1 * q.mm
        delta = 0.5 * q.mm
        await self.motor.set_position(position)
        await self.motor.move(delta)
        self.assertEqual(position + delta, await self.motor.get_position())


class TestContinuousLinearMotor(TestCase):

    def setUp(self):
        super(TestContinuousLinearMotor, self).setUp()
        self.motor = ContinuousLinearMotor()

    def test_set_position(self):
        position = 1 * q.mm
        self.motor.position = position
        self.assertEqual(position, self.motor.position)
        self.assertEqual(self.motor.state, 'standby')

    async def test_set_velocity(self):
        velocity = 1 * q.mm / q.s
        await self.motor.set_velocity(velocity)
        assert_almost_equal(velocity, await self.motor.get_velocity(), 0.1)
        self.assertEqual(await self.motor.get_state(), 'moving')
        await self.motor.stop()


class TestRotationMotor(TestCase):

    def setUp(self):
        super(TestRotationMotor, self).setUp()
        self.motor = RotationMotor()

    def test_set_position(self):
        position = 1 * q.deg
        self.motor.position = position
        self.assertEqual(position, self.motor.position)
        self.assertEqual(self.motor.state, 'standby')

    async def test_move(self):
        position = 1 * q.deg
        delta = 0.5 * q.deg
        await self.motor.set_position(position)
        await self.motor.move(delta)
        self.assertEqual(position + delta, await self.motor.get_position())
        self.assertEqual(await self.motor.get_state(), 'standby')


class TestContinuousRotationMotor(TestCase):

    def setUp(self):
        super(TestContinuousRotationMotor, self).setUp()
        self.motor = ContinuousRotationMotor()

    def test_set_position(self):
        position = 1 * q.deg
        self.motor.position = position
        self.assertEqual(position, self.motor.position)

    async def test_set_velocity(self):
        velocity = 1 * q.deg / q.s
        await self.motor.set_velocity(velocity)
        assert_almost_equal(velocity, await self.motor.get_velocity(), 0.1)
        self.assertEqual(await self.motor.get_state(), 'moving')
        await self.motor.stop()



#!/usr/bin/python
"""
Tango server for elmo_main.py-controller based motors
"""
import asyncio
import concert.writers
import numpy as np
import pickle
import time
import tango
import zmq
import zmq.asyncio
from concert.coroutines.base import start
from concert.helpers import PerformanceTracker
from concert.quantities import q
from concert.storage import DirectoryWalker
from tango import InfoIt, DebugIt
from PyTango.server import Device, attribute, DeviceMeta, pipe
from PyTango.server import AttrWriteType, command, device_property
from PyTango import DevState
from PyTango import ArgType, AttrDataFormat


async def coro():
    print('start')
    await asyncio.sleep(5)
    print('end')

    return 'done'


class DummyMotor(Device):
    """
    Device server for elmo_main.py-controller based
    """
    __metaclass__ = DeviceMeta
    volume = attribute(label="Volume", dtype=((np.float32,),),
            max_dim_x=1024, max_dim_y=1024,
                         access=AttrWriteType.READ,
                         fget="get_volume")

    endpoint = attribute(label="Endpoint", dtype=str,
                         access=AttrWriteType.READ_WRITE,
                         fget="get_endpoint", fset="set_endpoint")

    writer_class = attribute(label="Writerclass", dtype=str,
                         access=AttrWriteType.READ_WRITE,
                         fget="get_writer_class", fset="set_writer_class")

    dsetname = attribute(label="Dsetname", dtype=str,
                         access=AttrWriteType.READ_WRITE,
                         fget="get_dsetname", fset="set_dsetname")

    bytes_per_file = attribute(label="bytes_per_file", dtype=int,
                         access=AttrWriteType.READ_WRITE,
                         fget="get_bytes_per_file", fset="set_bytes_per_file")

    def __init__(self, a, b):
        self._context = zmq.asyncio.Context()
        self._socket = None
        self._task = None
        self._stop_at = None
        self._endpoint = None
        self._path = None
        # Walker
        self._writer_class = 'TiffWriter'
        self._dsetname = 'frame_{:>06}.tif'
        self._bytes_per_file = 2 ** 40
        super().__init__(a, b)

    async def init_device(self):
        """Inits device and communciation"""
        self.debug_stream("Init device")
        super().init_device()
        if self.get_state() == tango.DevState.RUNNING:
            self.debug_stream("Cancelling task: %s", self._task.cancel())
        # await self.disconnect_zmq()
        # print(self._socket, self._endpoint)
        if self._endpoint and not self._socket:
            await self.setup()
        self.set_state(tango.DevState.STANDBY)

    @DebugIt()
    @command()
    async def setup(self):
        if not self._endpoint:
            raise RuntimeError('Cannot connect when endpoint is not specified')

        if self._socket:
            self._socket.close()

        self._socket = self._context.socket(zmq.PULL)
        self._socket.connect(self._endpoint)
        

    @DebugIt()
    @command()
    async def teardown(self):
        if self._socket:
            self._socket.close()
            self._socket = None

    @InfoIt()
    async def get_endpoint(self):
        return self._endpoint

    @InfoIt(show_args=True)
    async def set_endpoint(self, endpoint):
        if self._endpoint == endpoint:
            return

        self._endpoint = endpoint
        await self.setup()

    @DebugIt()
    @command(dtype_in=str)
    def set_path(self, path):
        self._path = path

    @DebugIt()
    @command()
    def write_sequence(self):
        def callback(task):
            self.debug_stream('start callback, cancelled: %s, exception: %s',
                              task.cancelled(), task.exception())
            self.set_state(tango.DevState.STANDBY)
        walker = DirectoryWalker(
            writer=getattr(concert.writers, self._writer_class),
            dsetname=self._dsetname,
            bytes_per_file=self._bytes_per_file,
            root=self._path
        )
        self._stop_at = None
        self._task = start(self.consume(walker))
        self._task.add_done_callback(callback)
        self.set_state(tango.DevState.RUNNING)

    @DebugIt(show_args=True)
    @command(dtype_in=int)
    async def wait(self, num):
        self._stop_at = num
        if self._task is None:
            raise RuntimeError('Not started')

        try:
            await self._task
        finally:
            self._task = None
            self._stop_at = None

    async def consume(self, walker):
        with PerformanceTracker() as pt:
            total_bytes = await walker.write(self.subscribe())
            pt.size = total_bytes * q.B

    @InfoIt()
    def get_writer_class(self):
        return self._writer_class

    @InfoIt()
    def set_writer_class(self, writer_class):
        self._writer_class = writer_class

    @InfoIt()
    def get_dsetname(self):
        return self._dsetname

    @InfoIt()
    def set_dsetname(self, dsetname):
        self._dsetname = dsetname

    @InfoIt()
    def get_bytes_per_file(self):
        return self._bytes_per_file

    @InfoIt()
    def set_bytes_per_file(self, bytes_per_file):
        self._bytes_per_file = bytes_per_file

    @InfoIt()
    def get_volume(self):
        # import numpy as np
        return np.arange(25, dtype=np.float32).reshape(5, 5)
        # return [[1, 2, 3], [4, 5, 6.]]

    async def subscribe(self):
        i = 0

        while True:
            num_tries = 0
            while True:
                # We might get None if we are faster than the sender or the stopping count is set
                # after the stream ends
                image = await self._recv_array()
                if image is None:
                    num_tries += 1
                else:
                    break
                if self._stop_at is not None and i >= self._stop_at:
                    break
            if image is None:
                print(f'Stopping at i={i:>5} (tries={num_tries:>4}), stop at: {self._stop_at}')
                break
            yield image
            i += 1
            print(f'i={i:>5} (tries={num_tries:>4}), stop at: {self._stop_at}')

    async def _recv_array(self):
        import zmq

        try:
            md = await self._socket.recv_json(flags=zmq.NOBLOCK)
            msg = await self._socket.recv()
            array = np.frombuffer(msg, dtype=md['dtype'])

            return array.reshape(md['shape'])
        except zmq.Again:
            return None


def run_server():
    DummyMotor.run_server(
            args=['name', '-ORBendPoint', 'giop:tcp::1235', '-v4', '-nodb', '-dlist', 'a/b/c'],
            green_mode=tango.GreenMode.Asyncio
        )


if __name__ == "__main__":
    run_server()



from concert.tests import TestCase
from nose.plugins.attrib import attr


class TestIssue400(TestCase):

    @attr("skip-travis")
    def test_multiple_cameras(self):
        from concert.devices.cameras.uca import Camera

        cam_a = Camera('mock')
        cam_b = Camera('file')
        cam_c = Camera('mock')

        self.assertEqual("degree_value" in dir(cam_b), False)
        self.assertEqual("path" in dir(cam_c), False)



from concert.tests import TestCase, assert_almost_equal
from concert.devices.detectors.dummy import Detector


class TestDetector(TestCase):
    def setUp(self):
        super(TestDetector, self).setUp()
        self.detector = Detector()

    def test_pixel_size(self):
        pixel_width = self.detector.camera.sensor_pixel_width
        pixel_height = self.detector.camera.sensor_pixel_height

        assert_almost_equal(self.detector.pixel_width, pixel_width * self.detector.magnification)
        assert_almost_equal(self.detector.pixel_height, pixel_height * self.detector.magnification)



"""Shutter Dummy."""
from concert.base import transition
from concert.devices.shutters import base


class Shutter(base.Shutter):

    """A dummy shutter that can be opened and closed."""

    def __init__(self):
        super(Shutter, self).__init__()

    @transition(target='open')
    async def _open(self):
        pass

    @transition(target='closed')
    async def _close(self):
        pass



from concert.devices.lenschangers.dummy import LensChanger
from concert.tests import TestCase


class TestLensChanger(TestCase):

    def setUp(self):
        super(TestLensChanger, self).setUp()
        self.lens_changer = LensChanger()

    def test_set_objective(self):
        lens0 = "objective_10x"
        self.lens_changer.objective = lens0
        self.assertEqual(lens0, self.lens_changer.objective)
        lens1 = 'objective_5x'
        self.lens_changer.objective = lens1
        self.assertEqual(lens1, self.lens_changer.objective)



"""Storage implementations."""
import contextlib
import os
import logging
import tifffile
from logging import FileHandler, Formatter
from concert.coroutines.base import background, feed_queue
from concert.writers import TiffWriter


LOG = logging.getLogger(__name__)


def read_image(filename):
    """
    Read image from file with *filename*. The file type is detected
    automatically.
    """
    for ext, reader in list(READERS.items()):
        if filename.lower().endswith(ext):
            return reader(filename)

    raise ValueError("Unsupported file type")


def read_tiff(file_name):
    """Read tiff file from disk by :py:mod:`tifffile` module."""
    with tifffile.TiffFile(file_name) as f:
        return f.asarray(out='memmap')


READERS = {".tif": read_tiff,
           ".tiff": read_tiff}

try:
    import fabio

    def read_edf_via_fabio(filename):
        edf = fabio.edfimage.edfimage()
        edf.read(filename)
        return edf.data

    for ext in ('.edf', '.edf.gz'):
        READERS[ext] = read_edf_via_fabio
except ImportError:
    pass


def write_tiff(file_name, data):
    """
    The default TIFF writer which uses :py:mod:`tifffile` module.
    Return the written file name.
    """
    tifffile.imsave(file_name, data)

    return file_name


def write_libtiff(file_name, data):
    """Write a TIFF file using pylibtiff. Return the written file name."""
    from libtiff import TIFF

    tiff_file = TIFF.open(file_name, "w")
    try:
        tiff_file.write_image(data)
    finally:
        tiff_file.close()

    return file_name


def create_directory(directory, rights=0o0750):
    """Create *directory* and all paths along the way if necessary."""
    if not os.path.exists(directory):
        LOG.debug("Creating directory {}".format(directory))
        os.makedirs(directory, rights)


def write_images(pqueue, writer=TiffWriter, prefix="image_{:>05}.tif", start_index=0,
                 bytes_per_file=0):
    """
    write_images(pqueue, writer=TiffWriter, prefix="image_{:>05}.tif", start_index=0,
                 bytes_per_file=0)

    Write images on disk with specified *writer* and file name *prefix*. Write to one file until the
    *bytes_per_file* bytes has been written. If it is 0, then one file per image is created.
    *writer* is a subclass of :class:`.writers.ImageWriter`. *start_index* specifies the number in
    the first file name, e.g. for the default *prefix* and *start_index* 100, the first file name
    will be image_00100.tif. If *prefix* is not formattable images are appended to the filename
    specified by *prefix*.
    """
    im_writer = None
    file_index = 0
    written = 0
    dir_name = os.path.dirname(prefix)
    # If there is no formatting user wants just one file, in which case we append
    append = prefix.format(0) == prefix
    if append:
        im_writer = writer(prefix, bytes_per_file, append=True)

    if dir_name and not os.path.exists(dir_name):
        create_directory(dir_name)

    i = 0

    try:
        while True:
            image = pqueue.get().data
            if image is None:
                pqueue.task_done()
                break
            if not append and (not im_writer or written + image.nbytes > bytes_per_file):
                if im_writer:
                    im_writer.close()
                    LOG.debug('Writer "{}" closed'.format(prefix.format(start_index
                                                                        + file_index - 1)))
                im_writer = writer(prefix.format(start_index + file_index), bytes_per_file)
                file_index += 1
                written = 0
            im_writer.write(image)
            written += image.nbytes
            i += 1
            pqueue.task_done()
    finally:
        if im_writer:
            im_writer.close()
            LOG.debug('Writer "{}" closed'.format(prefix.format(start_index + file_index - 1)))


class Walker(object):

    """
    A Walker moves through an abstract hierarchy and allows to write data
    at a specific location.
    """

    def __init__(self, root, dsetname='frames', log=None, log_handler=None):
        """Constructor. *root* is the topmost level of the data structure."""
        self._root = root
        self._current = self._root
        self.dsetname = dsetname
        self._log = log
        self._log_handler = log_handler

        if self._log:
            self._log_handler.setLevel(logging.INFO)
            formatter = Formatter("[%(asctime)s] %(levelname)s: %(name)s: %(message)s")
            self._log_handler.setFormatter(formatter)
            self._log.addHandler(self._log_handler)

    def __del__(self):
        """Destructor."""
        if self._log:
            self._log_handler.close()
            self._log.removeHandler(self._log_handler)

    @contextlib.contextmanager
    def inside(self, name):
        self.descend(name)
        try:
            yield
        finally:
            self.ascend()

    def home(self):
        """Return to root."""
        self._current = self._root

    @property
    def current(self):
        """Return current position."""
        return self._current

    def exists(self, *paths):
        """Return True if path from current position specified by a list of *paths* exists."""
        raise NotImplementedError

    def descend(self, name):
        """Descend to *name* and return *self*."""
        self._descend(name)

        return self

    def ascend(self):
        """Ascend from current depth and return *self*."""
        self._ascend()

        return self

    def _descend(self, name):
        """Descend to *name*."""
        raise NotImplementedError

    def _ascend(self):
        """Ascend from current depth."""
        raise NotImplementedError

    @background
    async def write(self, producer, dsetname=None):
        """Coroutine for writing data set *dsetname*."""
        raise NotImplementedError


class DummyWalker(Walker):
    def __init__(self, root=''):
        super(DummyWalker, self).__init__(root)
        self._paths = set([])

    @property
    def paths(self):
        return self._paths

    def exists(self, *paths):
        return os.path.join(*paths) in self._paths

    def _descend(self, name):
        self._current = os.path.join(self._current, name)
        self._paths.add(self._current)

    def _ascend(self):
        if self._current != self._root:
            self._current = os.path.dirname(self._current)

    @background
    async def write(self, producer, dsetname=None):
        dsetname = dsetname or self.dsetname
        path = os.path.join(self._current, dsetname)

        i = 0
        async for item in producer:
            self._paths.add(os.path.join(path, str(i)))
            i += 1


class DirectoryWalker(Walker):

    """
    A DirectoryWalker moves through a file system and writes flat files using a
    specific filename template.
    """

    def __init__(self, writer=TiffWriter, dsetname='frame_{:>06}.tif', start_index=0,
                 bytes_per_file=0, root=None, log=None, log_name='experiment.log'):
        """
        Use *writer* to write data to files with filenames with a template from *dsetname*.
        *start_index* specifies the number in the first file name, e.g. for the default *dsetname*
        and *start_index* 100, the first file name will be frame_000100.tif.
        """
        if not root:
            root = os.getcwd()
        root = os.path.abspath(root)

        log_handler = None

        if log:
            create_directory(root)
            log_path = os.path.join(root, log_name)
            log_handler = FileHandler(log_path)

        super(DirectoryWalker, self).__init__(root, dsetname=dsetname,
                                              log=log, log_handler=log_handler)
        self._writer = writer
        self._bytes_per_file = bytes_per_file
        self._start_index = start_index

    def _descend(self, name):
        new = os.path.join(self._current, name)
        create_directory(new)
        self._current = new

    def _ascend(self):
        if self._current == self._root:
            raise StorageError("Cannot break out of `{}'.".format(self._root))

        self._current = os.path.dirname(self._current)

    def exists(self, *paths):
        """Check if *paths* exist."""
        return os.path.exists(os.path.join(self.current, *paths))

    @background
    async def write(self, producer, dsetname=None):
        dsetname = dsetname or self.dsetname

        if self._dset_exists(dsetname):
            dset_prefix = split_dsetformat(dsetname)
            dset_path = os.path.join(self.current, dset_prefix)
            raise StorageError("`{}' is not empty".format(dset_path))

        prefix = os.path.join(self._current, dsetname)

        return await feed_queue(producer, write_images, self._writer, prefix,
                                self._start_index, self._bytes_per_file)

    def _dset_exists(self, dsetname):
        """Check if *dsetname* exists on the current level."""
        bad = '{' not in dsetname

        try:
            dsetname.format(0)
        except ValueError:
            bad = True

        if bad:
            raise ValueError('dsetname `{}\' has wrong format'.format(dsetname))

        filenames = os.listdir(self._current)
        for name in filenames:
            if name.startswith(split_dsetformat(dsetname)):
                return True

        return False


def split_dsetformat(dsetname):
    """Strip *dsetname* off the formatting part wihch leaves us with the data set name."""
    return dsetname.split('{')[0]


class StorageError(Exception):
    """Exceptions related to logical issues with storage."""
    pass



"""A synchrotron beam line specific processes."""
import asyncio
import numpy as np
from concert.quantities import q
from concert.coroutines.base import background
from concert.imageprocessing import compute_rotation_axis, flat_correct
from concert.helpers import expects, Numeric
from concert.devices.motors.base import LinearMotor, RotationMotor
from concert.devices.cameras.base import Camera
from concert.devices.shutters.base import Shutter


@background
@expects(Camera, Shutter)
async def acquire_dark(camera, shutter):
    """Use *camera* and *shutter* to acquire a dark field."""
    if await shutter.get_state() != 'closed':
        await shutter.close()

    return await _grab_software_trigger(camera)


@background
@expects(Camera, Shutter, LinearMotor, Numeric(1, q.mm))
async def acquire_image_with_beam(camera, shutter, flat_motor, position):
    """Use *camera*, *shutter* and *flat_motor* to move the sample to the *position* and acquire an
    image.
    """
    coros = []
    if await shutter.get_state() != 'open':
        coros.append(shutter.open())
    coros.append(flat_motor.set_position(position))
    await asyncio.gather(*coros)

    return await _grab_software_trigger(camera)


@background
@expects(Camera, Shutter, LinearMotor, RotationMotor, Numeric(1, q.mm), Numeric(1, q.mm))
async def determine_rotation_axis(camera, shutter, flat_motor, rotation_motor, flat_position,
                                  radio_position):
    """Determine tomographic rotation axis using *camera*, *shutter*, *rotation_motor* and
    *flat_motor* devices.  *flat_position* is the position for making a flat field, *radio_position*
    is the position with the sample in the field of view.
    """
    dark = (await acquire_dark(camera, shutter)).astype(np.float32)
    flat = await acquire_image_with_beam(camera, shutter, flat_motor, flat_position)
    first = await acquire_image_with_beam(camera, shutter, flat_motor, radio_position)
    await rotation_motor.move(180 * q.deg)
    last = await acquire_image_with_beam(camera, shutter, flat_motor, radio_position)

    # flat correct
    flat_corr_first = flat_correct(first, flat, dark=dark)
    flat_corr_last = flat_correct(last, flat, dark=dark)

    return compute_rotation_axis(flat_corr_first, flat_corr_last)


async def _grab_software_trigger(camera):
    """Switch *camera* to software trigger source and take an image. The source is restored afterwards.
    """
    await camera['trigger_source'].stash()
    await camera.set_trigger_source(camera.trigger_sources.SOFTWARE)

    try:
        await camera.trigger()
        return await camera.grab()
    finally:
        await camera['trigger_source'].restore()



"""Add-ons for acquisitions are standalone extensions which can be applied to them. They operate on
the acquired data, e.g. write images to disk, do tomographic reconstruction etc.
"""
import logging
from concert.coroutines.base import async_generate
from concert.coroutines.sinks import Accumulate


LOG = logging.getLogger(__name__)


class Addon(object):

    """A base addon class. An addon can be attached, i.e. its functionality is applied to the
    specified *acquisitions* and detached.

    .. py:attribute:: acquisitions

    A list of :class:`~concert.experiments.base.Acquisition` objects. The addon attaches itself on
    construction.

    """

    def __init__(self, acquisitions):
        self.acquisitions = acquisitions
        self._attached = False
        self.attach()

    def attach(self):
        """Attach the addon to all acquisitions."""
        if self._attached:
            LOG.debug('Cannot attach an already attached Addon')
        else:
            self._attach()
            self._attached = True

    def detach(self):
        """Detach the addon from all acquisitions."""
        if self._attached:
            self._detach()
            self._attached = False
        else:
            LOG.debug('Cannot detach an unattached Addon')

    def _attach(self):
        """Attach implementation."""
        raise NotImplementedError

    def _detach(self):
        """Detach implementation."""
        raise NotImplementedError


class Consumer(Addon):

    """An addon which applies a specific coroutine-based consumer to acquisitions.

    .. py:attribute:: acquisitions

    a list of :class:`~concert.experiments.base.Acquisition` objects

    .. py:attribute:: consumer

    A callable which returns a coroutine which processes the incoming data from acquisitions

    """

    def __init__(self, acquisitions, consumer):
        self.consumer = consumer
        super(Consumer, self).__init__(acquisitions)

    def _attach(self):
        """Attach all acquisitions."""
        for acq in self.acquisitions:
            acq.consumers.append(self.consumer)

    def _detach(self):
        """Detach all acquisitions."""
        for acq in self.acquisitions:
            acq.consumers.remove(self.consumer)


class Accumulator(Addon):

    """An addon which accumulates data.

    .. py:attribute:: acquisitions

    a list of :class:`~concert.experiments.base.Acquisition` objects

    .. py:attribute:: shapes

    a list of shapes for different acquisitions

    .. py:attribute:: dtype

    the numpy data type
    """

    def __init__(self, acquisitions, shapes=None, dtype=None):
        self._accumulators = {}
        self._shapes = shapes
        self._dtype = dtype
        self.items = {}
        super(Accumulator, self).__init__(acquisitions)

    def _attach(self):
        """Attach all acquisitions."""
        shapes = (None,) * len(self.acquisitions) if self._shapes is None else self._shapes

        for i, acq in enumerate(self.acquisitions):
            self._accumulators[acq] = Accumulate(shape=shapes[i], dtype=self._dtype)
            self.items[acq] = self._accumulators[acq].items
            acq.consumers.append(self._accumulators[acq])

    def _detach(self):
        """Detach all acquisitions."""
        self.items = {}
        for acq in self.acquisitions:
            acq.consumers.remove(self._accumulators[acq])

        self._accumulators = {}


class ImageWriter(Addon):

    """An addon which writes images to disk.

    .. py:attribute:: acquisitions

    a list of :class:`~concert.experiments.base.Acquisition` objects

    .. py:attribute:: walker

    A :class:`~concert.storage.Walker` instance
    """

    def __init__(self, acquisitions, walker):
        self.walker = walker
        self._writers = {}
        super(ImageWriter, self).__init__(acquisitions)

    def _attach(self):
        """Attach all acquisitions."""
        for acq in self.acquisitions:
            self._writers[acq] = self._write_sequence(acq)
            acq.consumers.append(self._writers[acq])

    def _detach(self):
        """Detach all acquisitions."""
        for acq in self.acquisitions:
            acq.consumers.remove(self._writers[acq])
            del self._writers[acq]

    def _write_sequence(self, acquisition):
        """Wrap the walker and write data."""
        async def wrapped_writer(producer):
            """Returned wrapper."""
            try:
                self.walker.descend(acquisition.name)
                await self.walker.write(producer)
            finally:
                self.walker.ascend()

        return wrapped_writer


class OnlineReconstruction(Addon):
    def __init__(self, experiment, reco_args, do_normalization=True,
                 average_normalization=True, walker=None, slice_directory='online-slices'):
        from concert.ext.ufo import GeneralBackprojectManager

        self.experiment = experiment
        self.manager = GeneralBackprojectManager(reco_args,
                                                 average_normalization=average_normalization)
        self.walker = walker
        self.slice_directory = slice_directory
        self._consumers = {}
        self._do_normalization = do_normalization
        super().__init__(experiment.acquisitions)

    async def _reconstruct(self, producer):
        await self.manager.backproject(producer)
        if self.walker:
            with self.walker.inside(self.slice_directory):
                producer = async_generate(self.manager.volume)
                await self.walker.write(producer, dsetname='slice_{:>04}.tif')

    def _attach(self):
        if self._do_normalization:
            self._consumers[self.experiment.darks] = self.manager.update_darks
            self._consumers[self.experiment.flats] = self.manager.update_flats
        self._consumers[self.experiment.radios] = self._reconstruct

        for acq, consumer in self._consumers.items():
            acq.consumers.append(consumer)

    def _detach(self):
        for acq, consumer in list(self._consumers.items()):
            acq.consumers.remove(consumer)


class AddonError(Exception):

    """Addon errors."""

    pass


class OnlineReconstructionError(Exception):
    pass



"""Progress bar allows you to see the progress of long-term operations."""

import concert.config as cfg
import inspect


try:
    import tqdm
    from tqdm.asyncio import tqdm as atqdm
except ImportError:
    cfg.PROGRESS_BAR = False


def wrap_iterable(iterable, **kwargs):
    """Wrap *iterable* so that a progress bar will be shown on iteration."""
    if cfg.PROGRESS_BAR:
        if inspect.isasyncgen(iterable):
            iterable = atqdm(iterable, **kwargs)
        else:
            iterable = tqdm.tqdm(iterable, **kwargs)

    return iterable



"""
Imaging positioner which enables us to move around by specyfying distances
in the number of pixels.
"""
import numpy as np
from concert.coroutines.base import background
from concert.quantities import q
from concert.devices.positioners import base


class Positioner(base.Positioner):

    """
    A positioner which takes into account a detector with some pixel size.
    This way the user can specify the movement in pixels.
    """

    def __init__(self, axes, detector, position=None):
        super(Positioner, self).__init__(axes, position=position)
        self.detector = detector

    @background
    async def move(self, position):
        """Move by specified *position* which can be given in meters or pixels."""
        # Is there a better way to check units?
        if str(position.units) == 'pixel':
            mag = position.magnitude
            if mag[-1] != 0 and not np.isnan(mag[-1]):
                raise base.PositionerError('Cannot set \'z\' coordinate position in pixels.')

            physical_position = [0, 0, 0]
            width = (await self.detector.get_pixel_width()).to(q.m).magnitude
            height = (await self.detector.get_pixel_height()).to(q.m).magnitude
            physical_position[0] = position[0].magnitude * width
            physical_position[1] = position[1].magnitude * height
            physical_position *= q.m
        else:
            physical_position = position

        return await super(Positioner, self).move(physical_position)



import numpy as np
from concert import optimization
from concert.base import HardLimitError
from concert.quantities import q
from concert.tests import slow, assert_almost_equal, TestCase
from concert.measures import DummyGradientMeasure
from concert.devices.motors.dummy import LinearMotor
from concert.optimization import optimize_parameter


class TestDummyFocusingWithSoftLimits(TestCase):

    def setUp(self):
        super(TestDummyFocusingWithSoftLimits, self).setUp()
        self.motor = LinearMotor(position=50 * q.mm)
        self.motor['position'].lower = 25 * q.mm
        self.motor['position'].upper = 75 * q.mm
        self.halver_kwargs = {"initial_step": 10 * q.mm,
                              "max_iterations": 1000}

    @slow
    async def test_maximum_out_of_soft_limits_right(self):
        feedback = DummyGradientMeasure(self.motor['position'], 80 * q.mm, negative=True)
        await optimize_parameter(self.motor["position"], feedback,
                                 await self.motor.get_position(),
                                 optimization.halver,
                                 alg_kwargs=self.halver_kwargs)
        assert_almost_equal(await self.motor.get_position(), 75 * q.mm)

    @slow
    async def test_maximum_out_of_soft_limits_left(self):
        feedback = DummyGradientMeasure(self.motor['position'], 20 * q.mm, negative=True)
        await optimize_parameter(self.motor["position"], feedback,
                                 await self.motor.get_position(),
                                 optimization.halver,
                                 alg_kwargs=self.halver_kwargs)
        assert_almost_equal(await self.motor.get_position(), 25 * q.mm)


class TestDummyFocusing(TestCase):

    def setUp(self):
        super(TestDummyFocusing, self).setUp()
        self.motor = LinearMotor(upper_hard_limit=2000 * q.mm, lower_hard_limit=-2000 * q.mm)
        self.motor.position = 0 * q.mm
        self.feedback = DummyGradientMeasure(self.motor['position'],
                                             18.75 * q.mm, negative=True)
        self.epsilon = 1 * q.um
        self.position_eps = 1e-1 * q.mm
        self.gradient_cmp_eps = 1e-1
        self.halver_kwargs = {"initial_step": 10 * q.mm,
                              "max_iterations": 3000}

    async def run_optimization(self, initial_position=1 * q.mm):
        # Reset position of DummyMotor
        self.motor._position = 0 * q.mm
        await optimize_parameter(self.motor["position"], self.feedback,
                                 initial_position, optimization.halver,
                                 alg_kwargs=self.halver_kwargs)

    async def check(self, other):
        assert_almost_equal(await self.motor.get_position(), other, 0.1)

    @slow
    async def test_maximum_in_limits(self):
        # Disable Hard Limits of DummyMotor
        self.motor._upper_hard_limit = np.inf * q.mm
        self.motor._lower_hard_limit = -np.inf * q.mm

        await self.run_optimization()
        await self.check(self.feedback.max_position)

    @slow
    async def test_huge_step_in_limits(self):
        self.halver_kwargs["initial_step"] = 1000 * q.mm
        # Enable Hard Limits of DummyMotor
        self.motor._upper_hard_limit = 200 * q.mm
        self.motor._lower_hard_limit = -200 * q.mm

        with self.assertRaises(HardLimitError):
            await self.run_optimization()

    @slow
    async def test_maximum_out_of_limits_right(self):
        self.motor['position'].upper = 100 * q.mm
        self.feedback.max_position = (self.motor['position'].upper + 50 * q.mm)

        with self.assertRaises(HardLimitError):
            await self.run_optimization()

    @slow
    async def test_maximum_out_of_limits_left(self):
        self.motor['position'].lower = -100 * q.mm
        self.feedback.max_position = (self.motor['position'].lower - 50 * q.mm)

        with self.assertRaises(HardLimitError):
            await self.run_optimization()

    @slow
    async def test_identical_gradients(self):
        # Some gradient level reached and then by moving to another position
        # the same level is reached again, but global gradient maximum
        # is not at this motor position.

        # Disable Hard Limits of DummyMotor
        self.motor._upper_hard_limit = np.inf * q.mm
        self.motor._lower_hard_limit = -np.inf * q.mm

        self.halver_kwargs["initial_step"] = 10 * q.mm
        await self.run_optimization(initial_position=-0.00001 * q.mm)
        await self.check(self.feedback.max_position)



import numpy as np
from concert.coroutines.base import async_generate
from concert.quantities import q
from concert.devices.motors.dummy import ContinuousRotationMotor
from concert.imageprocessing import find_needle_tips
from concert.tests import slow, TestCase
from concert.tests.util.rotationaxis import SimulationCamera
from concert.processes.common import scan, align_rotation_axis
from concert.measures import rotation_axis


class TestRotationAxisMeasure(TestCase):

    def setUp(self):
        super(TestRotationAxisMeasure, self).setUp()
        self.x_motor = ContinuousRotationMotor()
        self.y_motor = ContinuousRotationMotor()
        self.z_motor = ContinuousRotationMotor()

        # The bigger the image size, the more images we need to determine
        # the center correctly.
        self.image_source = SimulationCamera(128, self.x_motor["position"],
                                             self.y_motor["position"],
                                             self.z_motor["position"])

        # Allow 1 px misalignment in y-direction.
        self.eps = np.arctan(2 / self.image_source.rotation_radius) * q.rad

    async def make_images(self, x_angle, z_angle, intervals=10):
        await self.x_motor.set_position(z_angle)
        await self.z_motor.set_position(x_angle)
        values = np.linspace(0, 2 * np.pi, intervals) * q.rad
        async for pair in scan(self.y_motor["position"], values, self.image_source.grab):
            yield pair[1]

    async def align_check(self, x_angle, z_angle):
        images = self.make_images(x_angle, z_angle)
        tips = await find_needle_tips(images)
        phi, psi = rotation_axis(tips)[:2]

        assert phi - x_angle < self.eps
        assert np.abs(psi) - np.abs(z_angle) < self.eps

    async def center_check(self, images):
        tips = await find_needle_tips(images)
        center = rotation_axis(tips)[2]

        assert np.abs(center[1] - self.image_source.ellipse_center[1]) < 2
        assert np.abs(center[0] - self.image_source.ellipse_center[0]) < 2

    @slow
    async def test_out_of_fov(self):
        images = np.random.normal(size=(10, self.image_source.size, self.image_source.size))
        with self.assertRaises(ValueError) as ctx:
            tips = await find_needle_tips(async_generate(images))
            rotation_axis(tips)

        self.assertEqual("No sample tip points found.", str(ctx.exception))

    @slow
    async def test_center_no_rotation(self):
        images = self.make_images(0 * q.deg, 0 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_only_x(self):
        self.image_source.scale = (3, 0.33, 3)
        images = self.make_images(17 * q.deg, 0 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_only_z(self):
        images = self.make_images(0 * q.deg, 11 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_positive(self):
        images = self.make_images(17 * q.deg, 11 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_negative_positive(self):
        images = self.make_images(-17 * q.deg, 11 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_positive_negative(self):
        images = self.make_images(17 * q.deg, -11 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_center_negative(self):
        images = self.make_images(-17 * q.deg, -11 * q.deg, intervals=15)
        await self.center_check(images)

    @slow
    async def test_only_x(self):
        """Only misaligned laterally."""
        await self.align_check(0 * q.deg, 0 * q.deg)

    @slow
    async def test_only_z(self):
        """Only misaligned in the beam direction."""
        await self.align_check(0 * q.deg, 11 * q.deg)

    @slow
    async def test_huge_x(self):
        self.image_source.scale = (3, 0.25, 3)
        await self.align_check(60 * q.deg, 11 * q.deg)

    @slow
    async def test_huge_z(self):
        self.image_source.scale = (3, 0.25, 3)
        await self.align_check(11 * q.deg, 60 * q.deg)

    @slow
    async def test_positive(self):
        await self.align_check(17 * q.deg, 11 * q.deg)

    @slow
    async def test_negative_positive(self):
        await self.align_check(-17 * q.deg, 11 * q.deg)

    @slow
    async def test_positive_negative(self):
        await self.align_check(17 * q.deg, -11 * q.deg)

    @slow
    async def test_negative(self):
        await self.align_check(-17 * q.deg, -11 * q.deg)

    @slow
    async def test_pitch_sgn(self):
        self.image_source.size = 512
        # Image acquisition inverts the contrast, so invert it here to get it right there
        await self.x_motor.set_position(10 * q.deg)
        await self.y_motor.set_position(0 * q.deg)
        await self.z_motor.set_position(0 * q.deg)
        eps = 0.1 * q.deg
        await align_rotation_axis(self.image_source, self.y_motor, x_motor=self.x_motor,
                                  z_motor=self.z_motor, initial_x_coeff=2 * q.dimensionless,
                                  metric_eps=eps)

        assert np.abs(await self.x_motor.get_position()) < eps



"""Image writers for uniform acces by :func:`.storage.write_images`"""


class ImageWriter(object):
    def __init__(self, filename, bytes_per_file, append=False):
        self._writer = None

    def write(self, image):
        raise NotImplementedError

    def close(self):
        self._writer.close()


class TiffWriter(ImageWriter):
    def __init__(self, filename, bytes_per_file, append=False):
        import tifffile
        # 2 ** 25 from tifffile
        self._writer = tifffile.TiffWriter(filename, append=append,
                                           bigtiff=bytes_per_file >= 2 ** 32 - 2 ** 25)

    def write(self, image):
        self._writer.save(image)


class LibTiffWriter(ImageWriter):
    def __init__(self, filename, bytes_per_file, append=False):
        from libtiff import TIFF
        mode = 'a' if append else 'w'
        if bytes_per_file >= 2 ** 31:
            mode += '8'
        self._writer = TIFF.open(filename, mode)

    def write(self, image):
        self._writer.write_image(image)



import os
from concert import __version__
from setuptools import setup, find_packages


# Install Bash completion script only if installation is run as root
data_files = []

if hasattr(os, 'geteuid') and os.geteuid() == 0:
    data_files = [('/etc/bash_completion.d', ['extras/completion/concert.sh'])]


setup(
    name='concert',
    # For tests it is 3.8, but let's be a bit relaxed
    python_requires='>=3.7',
    version=__version__,
    author='Matthias Vogelgesang',
    author_email='matthias.vogelgesang@kit.edu',
    url='http://github.com/ufo-kit/concert',
    license='LGPL',
    packages=find_packages(exclude=['*.tests']),
    scripts=['bin/concert',
             'bin/concert-server',
             'bin/concert-connect'],
    data_files=data_files,
    exclude_package_data={'': ['README.rst']},
    description="Lightweight beamline control system",
    long_description=open('README.rst').read(),
    install_requires=['numpy',
                      'prettytable',
                      'pyxdg',
                      'pint',
                      'scipy',
                      'tifffile'],
    tests_require=['nose',
                   'testfixtures'],
    test_suite='concert.tests',
)



import asyncio
import time
from itertools import product
from functools import reduce
import numpy as np
import logging
from concert.coroutines.base import background, broadcast
from concert.coroutines.sinks import Result
from concert.quantities import q
from concert.measures import rotation_axis
from concert.optimization import halver, optimize_parameter
from concert.imageprocessing import flat_correct, find_needle_tips
from concert.helpers import expects, is_iterable, Numeric
from concert.devices.motors.base import LinearMotor, RotationMotor
from concert.devices.shutters.base import Shutter
from concert.devices.cameras.base import Camera
from concert.progressbar import wrap_iterable


LOG = logging.getLogger(__name__)


async def scan(params, values, feedback, go_back=False):
    """Multi-dimensional scan of :class:`concert.base.Parameter` instances *params*, which are set
    to *values*. *feedback* is a coroutine function without parameters called after every iteration.
    If *go_back* is True, the original parameter values are restored at the end.

    If *params* is just one parameter and *values* is one list of values, perform a 1D scan. In this
    case tuples (x, y) are returned where x are the individual elements from the list of *values*
    and y = feedback() is called after every value setting.

    If *params* is a list of parameters and *values* is a list of lists, assign values[i] to
    params[i] and do a multi-dimensional scan, where last parameter changes the fastest (in other
    words a nested scan of all parameters, where *feedback* is called for every combination of
    parameter values. The combinations are obtained as a cartesian product of the *values*. For
    example, scanning camera exposure times and motor positions with values=[[1, 2] * q.s, [3, 5] *
    q.mm], would result in this::

        [((1 * q.s, 3 * q.mm), feedback()), ((1 * q.s, 5 * q.mm), feedback()),
         ((2 * q.s, 3 * q.mm), feedback()), ((2 * q.s, 5 * q.mm), feedback())]

    In general, for n parameters and lists of values, returned are tuples ((x_0, ..., x_{n-1}), y),
    where y = feedback() is called after every value setting (any parameter change). Parameter
    setting occurs in parallel, is waited for and then *feedback* is called.

    A simple 1D example::

        async for vector in scan(camera['exposure_time'], np.arange(1, 10, 1) * q.s, feedback):
            print(vector) # prints (1 * q.s, feedback()) and so on

    2D example::

        params = [camera['exposure_time'], motor['position']]
        values = [np.arange(1, 10, 1) * q.s, np.arange(5, 15, 2) * q.mm]
        async for vector in scan(params, values, feedback):
            print(vector) # prints ((1 * q.s, 5 * q.mm), feedback()) and so on

    """
    params = params if is_iterable(params) else [params]
    ndim = len(params)
    values = values if is_iterable(values[0]) else [values]

    if ndim > 1 and len(params) != len(values):
        raise RuntimeError
    num_iterations = reduce(lambda x, y: x * y, [len(vals) for vals in values])

    if go_back:
        for param in params:
            await param.stash()

    try:
        for vector in wrap_iterable(product(*values), total=num_iterations):
            setters = [params[i].set(vector[i]) for i in range(len(vector))]
            await asyncio.gather(*setters)
            yield (vector[0] if ndim == 1 else vector, await feedback())
    finally:
        if go_back:
            for param in params:
                await param.restore()


async def ascan(param, start, stop, step, feedback, go_back=False):
    """A convenience function to perform a 1D scan on parameter *param*, scan from *start* value to
    *stop* with *step*. *feedback* and *go_back* are the same as in the :func:`.scan`. This function
    just computes the values from *start*, *stop*, *step* and then calls :func:`.scan`::

        scan(param, values, feedback=feedback, go_back=go_back))
    """
    if start.units != stop.units or stop.units != step.units:
        raise RuntimeError
    region = np.arange(start.magnitude, stop.magnitude, step.magnitude) * start.units

    async for item in scan(param, region, feedback, go_back=go_back):
        yield item


async def dscan(param, delta, step, feedback, go_back=False):
    """A convenience function to perform a 1D scan on parameter *param*, scan from its current value
    to some *delta* with *step*. *feedback* and *go_back* are the same as in the :func:`.scan`. This
    function just computes the start and stop values and calls :func:`.ascan`::

        start = await param.get()
        ascan(param, start, start + delta, step, feedback, go_back=go_back)
    """
    start = await param.get()

    async for item in ascan(param, start, start + delta, step, feedback, go_back=go_back):
        yield item


@background
@expects(Camera, LinearMotor, measure=None, opt_kwargs=None,
         plot_callback=None, frame_callback=None)
async def focus(camera, motor, measure=np.std, opt_kwargs=None,
                plot_callback=None, frame_callback=None):
    """
    Focus *camera* by moving *motor*. *measure* is a callable that computes a
    scalar that has to be maximized from an image taken with *camera*.
    *opt_kwargs* are keyword arguments sent to the optimization algorithm.
    *plot_callback* is fed with y values from the optimization and
    *frame_callback* is a coroutine function fed with the incoming frames.

    This function is returning a future encapsulating the focusing event. Note,
    that the camera is stopped from recording as soon as the optimal position
    is found.
    """
    if opt_kwargs is None:
        opt_kwargs = {'initial_step': 0.5 * q.mm,
                      'epsilon': 1e-2 * q.mm}

    async def get_measure():
        await camera.trigger()
        frame = await camera.grab()
        if frame_callback:
            await frame_callback(frame)
        return - measure(frame)

    async def filter_optimization(xy):
        """
        Filter the optimization's (x, y) subresults to only the y part,
        otherwise the the plot update is not lucid.
        """
        if plot_callback:
            await plot_callback(xy[1])

    await camera['trigger_source'].stash()
    await camera.set_trigger_source(camera.trigger_sources.SOFTWARE)

    try:
        async with camera.recording():
            await optimize_parameter(motor['position'], get_measure,
                                     await motor.get_position(),
                                     halver, alg_kwargs=opt_kwargs,
                                     callback=filter_optimization)
    finally:
        await camera['trigger_source'].restore()


@expects(Camera, RotationMotor, num_frames=Numeric(1), shutter=Shutter,
         flat_motor=LinearMotor, flat_position=Numeric(1, q.m), y_0=Numeric(1),
         y_1=Numeric(1))
async def acquire_frames_360(camera, rotation_motor, num_frames, shutter=None, flat_motor=None,
                             flat_position=None, y_0=0, y_1=None):
    """
    acquire_frames_360(camera, rotation_motor, num_frames, shutter=None, flat_motor=None,
                       flat_position=None, y_0=0, y_1=None)

    Acquire frames around a circle.
    """
    flat = None
    if await camera.get_state() == 'recording':
        await camera.stop_recording()
    await camera['trigger_source'].stash()
    await camera.set_trigger_source(camera.trigger_sources.SOFTWARE)

    try:
        async with camera.recording():
            if shutter:
                if await shutter.get_state() != 'closed':
                    await shutter.close()
                await camera.trigger()
                dark = (await camera.grab())[y_0:y_1]
                await shutter.open()
            if flat_motor and flat_position is not None:
                radio_position = await flat_motor.get_position()
                await flat_motor.set_position(flat_position)
                await camera.trigger()
                flat = (await camera.grab())[y_0:y_1]
                await flat_motor.set_position(radio_position)
            for i in range(num_frames):
                await rotation_motor.move(2 * np.pi / num_frames * q.rad)
                await camera.trigger()
                frame = (await camera.grab())[y_0:y_1].astype(np.float)
                if flat is not None:
                    frame = flat_correct(frame, flat, dark=dark)
                    frame = np.nan_to_num(-np.log(frame))
                    # Huge numbers can also cause trouble
                    frame[np.abs(frame) > 1e6] = 0
                else:
                    frame = frame.max() - frame
                yield frame
    finally:
        # No side effects
        await camera['trigger_source'].restore()


@background
@expects(Camera, RotationMotor, x_motor=RotationMotor, z_motor=RotationMotor,
         get_ellipse_points=find_needle_tips, num_frames=Numeric(1), metric_eps=Numeric(1, q.deg),
         position_eps=Numeric(1, q.deg), max_iterations=Numeric(1),
         initial_x_coeff=Numeric(1, q.dimensionless), initial_z_coeff=Numeric(1, q.dimensionless),
         shutter=Shutter, flat_motor=LinearMotor, flat_position=Numeric(1, q.m),
         y_0=Numeric(1), y_1=Numeric(1), get_ellipse_points_kwargs=None, frame_consumer=None)
async def align_rotation_axis(camera, rotation_motor, x_motor=None, z_motor=None,
                              get_ellipse_points=find_needle_tips, num_frames=10, metric_eps=None,
                              position_eps=0.1 * q.deg, max_iterations=5,
                              initial_x_coeff=1 * q.dimensionless,
                              initial_z_coeff=1 * q.dimensionless,
                              shutter=None, flat_motor=None, flat_position=None, y_0=0, y_1=None,
                              get_ellipse_points_kwargs=None, frame_consumer=None):
    """
    align_rotation_axis(camera, rotation_motor, x_motor=None, z_motor=None,
    get_ellipse_points=find_needle_tips, num_frames=10, metric_eps=None,
    position_eps=0.1 * q.deg, max_iterations=5, initial_x_coeff=1 * q.dimensionless,
    initial_z_coeff=1 * q.dimensionless, shutter=None, flat_motor=None, flat_position=None,
    y_0=0, y_1=None, get_ellipse_points_kwargs=None, frame_consumer=None)

    Align rotation axis. *camera* is used to obtain frames, *rotation_motor* rotates the sample
    around the tomographic axis of rotation, *x_motor* turns the sample around x-axis, *z_motor*
    turns the sample around z-axis.

    *get_ellipse_points* is a function with one positional argument, a set of images. It computes
    the ellipse points from the sample positions as it rotates around the tomographic axis.  You can
    use e.g. :func:`concert.imageprocessing.find_needle_tips` and
    :func:`concert.imageprocessing.find_sphere_centers` to extract the ellipse points from needle
    tips or sphere centers. You can pass additional keyword arguments to the *get_ellipse_points*
    function in the *get_ellipse_points_kwargs* dictionary.

    *num_frames* defines how many frames are acquired and passed to the *measure*. *metric_eps* is
    the metric threshold for stopping the procedure. If not specified, it is calculated
    automatically to not exceed 0.5 pixels vertically. If *max_iterations* is reached the procedure
    stops as well. *initial_[x|z]_coeff* is the coefficient applied` to the motor motion for the
    first iteration. If we move the camera instead of the rotation stage, it is often necessary to
    acquire fresh flat fields. In order to make an up-to-date flat correction, specify *shutter* if
    you want fresh dark fields and specify *flat_motor* and *flat_position* to acquire flat fields.
    Crop acquired images to *y_0* and *y_1*. *frame_consumer* is a coroutine function which will be
    fed with all acquired frames.

    The procedure finishes when it finds the minimum angle between an ellipse extracted from the
    sample movement and respective axes or the found angle drops below *metric_eps*. The axis of
    rotation after the procedure is (0,1,0), which is the direction perpendicular to the beam
    direction and the lateral direction. *x_motor* and *z_motor* do not have to move exactly by the
    computed angles but their relative motion must be linear with respect to computed angles (e.g.
    if the motors operate with steps it is fine, also rotation direction does not need to be known).
    """
    if get_ellipse_points_kwargs is None:
        get_ellipse_points_kwargs = {}

    if not x_motor and not z_motor:
        raise ProcessError("At least one of the x, z motors must be given")

    async def make_step(i, motor, position_last, angle_last, angle_current, initial_coeff,
                        rotation_type):
        cur_pos = await motor.get_position()
        LOG.debug("%s: i: %d, last angle: %s, angle: %s, last position: %s, position: %s",
                  rotation_type, i, angle_last.to(q.deg), angle_current.to(q.deg),
                  position_last.to(q.deg), cur_pos.to(q.deg))
        if i > 0:
            # Assume linear mapping between the computed angles and motor motion
            if angle_current == angle_last:
                coeff = 0 * q.dimensionless
            else:
                coeff = (cur_pos - position_last) / (angle_current - angle_last)
        else:
            coeff = initial_coeff
        position_last = cur_pos
        angle_last = angle_current

        # Move relative, i.e. if *angle_current* should go to 0, then we need to move in the
        # other direction with *coeff* applied
        LOG.debug("%s coeff: %s, Next position: %s", rotation_type, coeff.to_base_units(),
                  (cur_pos - coeff * angle_current).to(q.deg))
        await motor.move(-coeff * angle_current)

        return (position_last, angle_last)

    async def go_to_best_index(motor, history):
        positions, angles = list(zip(*history))
        best_index = np.argmin(np.abs([angle.to_base_units().magnitude for angle in angles]))
        LOG.debug("Best iteration: %d, position: %s, angle: %s",
                  best_index, positions[best_index].to(q.deg), angles[best_index].to(q.deg))
        await motor.set_position(positions[best_index])

    async def extract_points(producer):
        return await get_ellipse_points(producer, **get_ellipse_points_kwargs)

    roll_history = []
    pitch_history = []
    center = None

    if z_motor:
        roll_angle_last = 0 * q.deg
        roll_position_last = await z_motor.get_position()
        roll_continue = True
    if x_motor:
        pitch_angle_last = 0 * q.deg
        pitch_position_last = await x_motor.get_position()
        pitch_continue = True

    frames_result = Result()
    for i in range(max_iterations):
        acq_consumers = [extract_points, frames_result]
        if frame_consumer:
            acq_consumers.append(frame_consumer)
        tips_start = time.perf_counter()
        frame_producer = acquire_frames_360(camera, rotation_motor, num_frames, shutter=shutter,
                                            flat_motor=flat_motor, flat_position=flat_position,
                                            y_0=y_0, y_1=y_1)

        coros = broadcast(frame_producer, *acq_consumers)
        try:
            tips = (await asyncio.gather(*coros))[1]
        except Exception as tips_exc:
            raise ProcessError('Error finding reference points') from tips_exc
        LOG.debug('Found %d points in %g s', len(tips), time.perf_counter() - tips_start)
        roll_angle_current, pitch_angle_current, center = rotation_axis(tips)
        coros = []
        x_coro = z_coro = None
        if metric_eps is None:
            metric_eps = np.rad2deg(np.arctan(1 / frames_result.result.shape[1])) * q.deg
            LOG.debug('Automatically computed metric epsilon: %s', metric_eps)

        if z_motor and roll_continue:
            z_pos = await z_motor.get_position()
            roll_history.append((z_pos, roll_angle_current))
            if (np.abs(roll_angle_current) >= metric_eps
                    and (np.abs(roll_position_last - z_pos) >= position_eps or i == 0)):
                z_coro = make_step(i, z_motor, roll_position_last, roll_angle_last,
                                   roll_angle_current, initial_z_coeff, 'roll')
                coros.append(z_coro)
            else:
                LOG.debug("Roll epsilon reached")
                roll_continue = False
        if x_motor and pitch_continue:
            x_pos = await x_motor.get_position()
            pitch_history.append((x_pos, pitch_angle_current))
            if (np.abs(pitch_angle_current) >= metric_eps
                    and (np.abs(pitch_position_last - x_pos) >= position_eps or i == 0)):
                x_coro = make_step(i, x_motor, pitch_position_last, pitch_angle_last,
                                   pitch_angle_current, initial_x_coeff, 'pitch')
                coros.append(x_coro)
            else:
                LOG.debug("Pitch epsilon reached")
                pitch_continue = False

        if not coros:
            # If there are no coros the motors have reached positions at which the computed
            # angles are below threshold
            break

        step_results = await asyncio.gather(*coros)
        if x_coro:
            # Regardless from x_coro and z_coro to be present, x_coro is always added last, so pop
            # it first
            pitch_position_last, pitch_angle_last = step_results.pop()
        if z_coro:
            roll_position_last, roll_angle_last = step_results.pop()

    if i == max_iterations - 1:
        LOG.info('Maximum iterations reached')

    # Move to the best known position
    coros = []
    if z_motor:
        coros.append(go_to_best_index(z_motor, roll_history))
    if x_motor:
        coros.append(go_to_best_index(x_motor, pitch_history))
    await asyncio.gather(*coros)

    return (roll_history, pitch_history, center)


class ProcessError(Exception):

    """
    Exception raised by a process when something goes wrong with the procedure
    it tries to accomplish, e.g. cannot focus, cannot align rotation axis, etc.

    """

    pass



"""Imaging experiments usually conducted at synchrotrons."""
import numpy as np
from concert.quantities import q


async def frames(num_frames, camera, callback=None):
    """
    A generator which takes *num_frames* using *camera*. *callback* is called
    after every taken frame.
    """
    if await camera.get_state() == 'recording':
        await camera.stop_recording()

    await camera['trigger_source'].stash()
    await camera.set_trigger_source(camera.trigger_sources.SOFTWARE)

    try:
        async with camera.recording():
            for i in range(num_frames):
                await camera.trigger()
                yield await camera.grab()
                if callback:
                    await callback()
    finally:
        await camera['trigger_source'].restore()


def tomo_angular_step(frame_width):
    """
    Get the angular step required for tomography so that every pixel of the frame
    rotates no more than one pixel per rotation step. *frame_width* is frame size in
    the direction perpendicular to the axis of rotation.
    """
    return np.arctan(2 / frame_width.magnitude) * q.rad


def tomo_projections_number(frame_width):
    """
    Get the minimum number of projections required by a tomographic scan in
    order to provide enough data points for every distance from the axis of
    rotation. The minimum angular step is
    considered to be needed smaller than one pixel in the direction
    perpendicular to the axis of rotation. The number of pixels in this
    direction is given by *frame_width*.
    """
    return int(np.ceil(np.pi / tomo_angular_step(frame_width)))


def tomo_max_speed(frame_width, frame_rate):
    """
    Get the maximum rotation speed which introduces motion blur less than one
    pixel. *frame_width* is the width of the frame in the direction
    perpendicular to the rotation and *frame_rate* defines the time required
    for recording one frame.

    _Note:_ frame rate is required instead of exposure time because the
    exposure time is usually shorter due to the camera chip readout time.
    We need to make sure that by the next exposure the sample hasn't moved
    more than one pixel from the previous frame, thus we need to take into
    account the whole frame taking procedure (exposure + readout).
    """
    return tomo_angular_step(frame_width) * frame_rate



"""
Optimization is a procedure to iteratively find the best possible match
to

.. math::
    y = f(x).

This module provides execution routines and algorithms for optimization.
"""
import logging
from functools import wraps
from concert.coroutines.base import background, run_in_executor, run_in_loop_thread_blocking
from concert.quantities import q


LOG = logging.getLogger(__name__)


@background
async def optimize(function, x_0, algorithm, alg_args=(), alg_kwargs=None,
                   callback=None):
    """
    optimize(function, x_0, algorithm, alg_args=(), alg_kwargs=None, callback=None)

    Optimize y = await *function* (x), so *function* must be a coroutine function. *x_0* is the
    initial guess. *algorithm* is the optimization algorithm to be used::

        algorithm(x_0, *alg_args, **alg_kwargs)

    *callback* is a callable called with all the (x, y) values as they are obtained.
    """
    alg_kwargs = {} if alg_kwargs is None else alg_kwargs
    data = []

    async def evaluate(x_val):
        """Execute y = f(*x*), save (x, y) pair and return y."""
        result = await function(x_val)
        pair = x_val, result
        if callback:
            await callback(pair)
        data.append(pair)

        return result

    result = await algorithm(evaluate, x_0, *alg_args, **alg_kwargs)

    return result, data


@background
async def optimize_parameter(parameter, feedback, x_0, algorithm, alg_args=(),
                             alg_kwargs=None, callback=None):
    """
    Optimize *parameter* and use the *feedback* (a coroutine function)
    as a result. Other arguments are the same as by :func:`optimize`.
    The function to be optimized is determined as follows::

        await parameter.set(x)
        y = await feedback()

    *callback* is the same as by :func:`optimize`.
    """
    async def function(x_val):
        """
        Create a function from setting a parameter and getting a feedback.
        """
        # Do not go out of the limits
        if parameter.lower is not None:
            if parameter.lower is not None and x_val < parameter.lower:
                x_val = parameter.lower
        if parameter.upper is not None:
            if parameter.upper is not None and x_val > parameter.upper:
                x_val = parameter.upper
        await parameter.set(x_val)

        return await feedback()

    return await optimize(function, x_0, algorithm, alg_args=alg_args,
                          alg_kwargs=alg_kwargs, callback=callback)


@background
async def halver(function, x_0, initial_step=None, epsilon=None,
                 max_iterations=100):
    """
    Halving the interval, evaluate *function* based on *param*. Use
    *initial_step*, *epsilon* precision and *max_iterations*.
    """
    if initial_step is None:
        # Figure out the step based on x_0 units (take one in the given unit)
        step = q.Quantity(1, x_0.units)
    else:
        step = initial_step
    if epsilon is None:
        epsilon = q.Quantity(1e-4, x_0.units)

    direction = 1
    i = 0
    last_x = x_0

    y_0 = await function(x_0)
    # Remember the best x and y
    best = x_0, y_0

    def turn(direction, step):
        """Turn to opposite direction and reduce the step by half."""
        return -direction, step / 2

    def move(x_0, direction, step):
        """Move to a *direction* by a *step*."""
        return x_0 + direction * step

    x_0 = move(x_0, direction, step)

    while i < max_iterations:
        y_1 = await function(x_0)

        if step < epsilon:
            break

        if y_1 >= y_0:
            # Worse, change direction and move to the half of the last
            # good x and the new x.
            direction, step = turn(direction, step)
            x_0 = (x_0 + last_x) / 2
        else:
            # OK, move forward.
            if y_1 < best[1]:
                # The new y is better then the so far obtained one
                best = x_0, y_1
            last_x = x_0
            x_0 = move(x_0, direction, step)

        y_0 = y_1
        i += 1

    # Apply the best known value
    await function(best[0])

    return best[0]


def _quantized(strip_func):
    """
    Decorator to quantize a *function* which does not take units into
    account. Strips intermediate results based on *strip_func* in order
    to fit *function*'s first parameter signature.
    """
    @wraps(strip_func)
    def stripped(function):
        @wraps(function)
        async def wrapper(eval_func, x_0, *args, **kwargs):
            def q_func(x):
                """We need to run the objective function in a separate thread and loop because we
                have been called from a running event loop and have things to do in an event loop
                and can't use the running one.
                """
                coro = eval_func(q.Quantity(strip_func(x), x_0.units))
                return run_in_loop_thread_blocking(coro)

            def to_run_in_exec():
                """Run the optimization *function* in a pool so that we comply with the async nature
                of concert.
                """
                return function(q_func, x_0.magnitude, *args, **kwargs)

            dim_less = await run_in_executor(to_run_in_exec)
            return q.Quantity(dim_less, x_0.units)
        return wrapper

    return stripped


@_quantized(lambda x: x[0])
def down_hill(function, x_0, **kwargs):
    """
    down_hill(function, x_0, **kwargs)

    Downhill simplex algorithm from :py:func:`scipy.optimize.fmin`.
    Please refer to the scipy function for additional arguments information.
    """
    from scipy import optimize as scipy_optimize

    return scipy_optimize.fmin(function, x_0, disp=0, **kwargs)[0]


@_quantized(lambda x: x[0])
def powell(function, x_0, **kwargs):
    """
    powell(function, x_0, **kwargs)

    Powell's algorithm from :py:func:`scipy.optimize.fmin_powell`.
    Please refer to the scipy function for additional arguments information.
    """
    from scipy import optimize as scipy_optimize

    return scipy_optimize.fmin_powell(function, x_0, disp=0, **kwargs)


@_quantized(lambda x: x[0])
def nonlinear_conjugate(function, x_0, **kwargs):
    """
    nonlinear_conjugate(function, x_0, **kwargs)

    Nonlinear conjugate gradient algorithm from
    :py:func:`scipy.optimize.fmin_cg`.
    Please refer to the scipy function for additional arguments information.
    """
    from scipy import optimize as scipy_optimize

    return scipy_optimize.fmin_cg(function, x_0, disp=0, **kwargs)[0]


@_quantized(lambda x: x[0])
def bfgs(function, x_0, **kwargs):
    """
    bfgs(function, x_0, **kwargs)

    Broyde-Fletcher-Goldfarb-Shanno (BFGS) algorithm from
    :py:func:`scipy.optimize.fmin_bfgs`.
    Please refer to the scipy function for additional arguments information.
    """
    from scipy import optimize as scipy_optimize

    return scipy_optimize.fmin_bfgs(function, x_0, disp=0, **kwargs)[0]


@_quantized(lambda x: x[0])
def least_squares(function, x_0, **kwargs):
    """
    least_squares(function, x_0, **kwargs)

    Least squares algorithm from :py:func:`scipy.optimize.leastsq`.
    Please refer to the scipy function for additional arguments information.
    """
    from scipy import optimize as scipy_optimize

    return scipy_optimize.leastsq(function, x_0, **kwargs)[0][0]



import sys
import concert.session.management as cs
from concert.helpers import Command, Bunch


class SpyderCommand(Command):
    """Start session using Spyder."""
    def __init__(self):
        opts = {'session': {'type': str}}
        super(SpyderCommand, self).__init__('spyder', opts)

    def run(self, session):
        from spyderlib import spyder
        from spyderlib.config import CONF

        cs.exit_if_not_exists(session)
        app = spyder.initialize()

        # This should come from our command line parser, however, Spyder does
        # not provide a way to get the argument parser but only the parsed
        # arguments.
        opts = {'working_directory': cs.path(),
                'debug': False,
                'profile': False,
                'multithreaded': False,
                'light': False,
                'new_instance': True}

        # The python executable is set explicitly here, because Spyder tends
        # not to pick up the one used in a virtualenv. This should be set in a
        # profile in order to not mess with the user's settings.
        CONF.set('console', 'pythonexecutable', sys.executable)

        main = spyder.MainWindow(Bunch(opts))
        main.setup()
        main.show()
        main.open_file(cs.path(session))

        app.exec_()



from concert.base import State, FSMError
from concert.devices.base import Device
from concert.tests import TestCase


class BadDevice(Device):

    state = State()


class ImplicitSoftwareDevice(Device):

    state = State(default='standby')


class CustomizedDevice(Device):

    async def state_getter(self):
        return 'custom'

    state = State(fget=state_getter)


class RealDevice(Device):

    state = State()

    def __init__(self):
        super(RealDevice, self).__init__()
        self._state = 'standby'

    def change_state(self):
        self._state = 'moved'

    async def _get_state(self):
        return self._state


class TestState(TestCase):

    async def test_bad(self):
        device = BadDevice()
        with self.assertRaises(FSMError):
            await device['state'].get()

    def test_real_device(self):
        device = RealDevice()
        self.assertEqual(device.state, 'standby')
        device.change_state()
        self.assertEqual(device.state, 'moved')

    def test_implicit(self):
        device = ImplicitSoftwareDevice()
        self.assertEqual(device.state, 'standby')

    def test_customized(self):
        device = CustomizedDevice()
        self.assertEqual(device.state, 'custom')



"""Handle session management.

A session is an ordinary Python module that is stored in a per-user
directory."""
import os
import sys
import imp
import shutil

_CACHED_PATH = None

_SESSION_TEMPLATE = """\"\"\"# *{doc}* session manual

## Usage

## Notes
\"\"\"

import logging
from inspect import iscoroutinefunction
import concert
concert.require("{version}")

from concert.quantities import q
from concert.session.utils import ddoc, dstate, pdoc, code_of

LOG = logging.getLogger(__name__)
"""


def path(session=None):
    """
    Get absolute path of *session* module or base path if *session* is None.
    """
    global _CACHED_PATH

    if _CACHED_PATH is None:
        env = os.environ
        if "VIRTUAL_ENV" in env:
            env["XDG_DATA_HOME"] = os.path.join(env["VIRTUAL_ENV"], "share")

        import xdg.BaseDirectory
        _CACHED_PATH = xdg.BaseDirectory.save_data_path('concert')

    if session is None:
        return _CACHED_PATH

    return os.path.join(_CACHED_PATH, session + '.py')


def logfile_path():
    return os.path.join(path(), 'concert.log')


def create(session, imports=()):
    """Create a template with *session* name and write it.

    For each name in *imports* try to load it and insert `from
    concert.processes.name import *` into the session file.

    .. note:: This will *always* overwrite session.
    """
    from concert import get_canonical_version
    template = _SESSION_TEMPLATE.format(version=get_canonical_version(), doc=session)

    def _module_exists(module_name):
        try:
            __import__(module_name)
            return True
        except ImportError:
            return False

    for module in imports:
        module_name = 'concert.processes.{0}'.format(module)

        if _module_exists(module_name):
            template += 'from {0} import *'.format(module_name)
        else:
            print("{0} not found.".format(module_name))

    if not os.path.exists(path()):
        os.mkdir(path())

    with open(path(session), 'w') as session_file:
        session_file.write(template)


def remove(session):
    """Remove a *session*."""
    if exists(session):
        os.unlink(path(session))


def move(source, target):
    """Move *source* to *target*."""
    os.rename(path(source), path(target))


def copy(source, target):
    """Copy *source* to *target*."""
    shutil.copy(path(source), path(target))


def load(session, from_file=False):
    """Load *session* and return the module."""
    if not from_file:
        return imp.load_source(session, path(session))

    return imp.load_source('somename', os.path.abspath(session))


def get_existing():
    """Get all existing session names."""
    sessions = [f for f in os.listdir(path()) if f.endswith('.py')]
    return [os.path.splitext(f)[0] for f in sessions]


def exists(session):
    """Check if *session* already exists."""
    return os.access(path(session), os.R_OK)


def exit_if_not_exists(session):
    """Exit if *session* does not exist with a message."""
    if not exists(session):
        message = "Session `{0}' does not exist. Run `concert init {0}' first."
        sys.exit(message.format(session))



import numpy as np
from concert.coroutines.base import async_generate
from concert.devices.motors.dummy import ContinuousRotationMotor
from concert.quantities import q
from concert.imageprocessing import (compute_rotation_axis, normalize, find_sphere_centers)
from concert.measures import rotation_axis
from concert.tests import suppressed_logging, slow, assert_almost_equal, TestCase
from concert.tests.util.rotationaxis import SimulationCamera


@slow
@suppressed_logging
def test_rotation_axis():
    """
    Test tomographic rotation axis finding. The sample is a triangle which
    is offcentered by different values.
    """
    def triangle(n, width, position, left=True):
        tr = np.zeros((width, width), dtype=np.float)
        extended = np.zeros((n, n), dtype=np.float)

        indices = np.tril_indices(width)
        tr[indices] = 1
        if not left:
            tr = tr[:, ::-1]
        extended[n // 2 - width // 2: n // 2 + width // 2, position:position + width] = tr

        return extended

    def test_axis(n, width, left_position, right_position):
        left = triangle(n, width, left_position)
        right = triangle(n, width, right_position, left=False)

        center = compute_rotation_axis(left, right)
        truth = (left_position + right_position + width) / 2 * q.px
        assert_almost_equal(center, truth)

    n = 128
    width = 32

    # Axis is exactly in the middle
    test_axis(n, width, n // 2 - width, n // 2)

    # Axis is to the left from the center
    test_axis(n, width, 8, n // 2 + width // 4)

    # Axis is to the right from the center
    test_axis(n, width, 8, n - width)


@suppressed_logging
def test_rescale():
    array = np.array([-2.5, -1, 14, 100])

    def run_test(minimum, maximum):
        conversion = (array.max() - array.min()) / (maximum - minimum)
        normed = normalize(array, minimum=minimum, maximum=maximum)

        np.testing.assert_almost_equal(np.gradient(array), np.gradient(normed) * conversion)
        np.testing.assert_almost_equal(normed[0], minimum)
        np.testing.assert_almost_equal(normed[-1], maximum)

    run_test(0, 1)
    run_test(-10, 47.5)


@slow
class TestSphereSegmentation(TestCase):

    def setUp(self):
        super(TestSphereSegmentation, self).setUp()
        self.x_motor = ContinuousRotationMotor()
        self.y_motor = ContinuousRotationMotor()
        self.z_motor = ContinuousRotationMotor()

        self.x_motor.position = 13 * q.deg
        self.y_motor.position = 0 * q.deg
        self.z_motor.position = -7 * q.deg

        self.camera = SimulationCamera(128, self.x_motor["position"],
                                       self.y_motor["position"],
                                       self.z_motor["position"],
                                       scales=(0.5, 0.5, 0.5),
                                       y_position=0)

    async def acquire_frames(self, num_frames=10):
        da = 360 * q.degree / num_frames
        images = []
        centers = []

        for i in range(10):
            await self.y_motor.move(da)
            images.append(await self.camera.grab())
            centers.append(self.camera.ellipsoid_center)

        return (images, centers)

    def check(self, gt, measured):
        assert np.all(np.abs(np.fliplr(gt) - measured) < 1)

    async def test_sphere_fully_inside(self):
        (frames, gt) = await self.acquire_frames()
        centers = await find_sphere_centers(async_generate(frames))
        self.check(gt, centers)

    async def test_sphere_some_partially_outside(self):
        self.camera.rotation_radius = self.camera.size // 2
        (frames, gt) = await self.acquire_frames()
        centers = await find_sphere_centers(async_generate(frames))
        self.check(gt, centers)

    async def test_sphere_all_partially_outside(self):
        self.camera.size = 128
        self.camera.rotation_radius = self.camera.size // 2
        self.camera.scale = (.25, .25, .25)
        frames = (await self.acquire_frames())[0]
        centers = await find_sphere_centers(async_generate(frames), correlation_threshold=0.9)
        roll, pitch = rotation_axis(centers)[:2]

        # No sphere is completely in the FOV, filter the ones with low correlation coefficient and
        # at least coarsely close match should be found to the ellipse
        assert np.abs(roll - await self.z_motor.get_position()) < 1 * q.deg
        assert np.abs(pitch - await self.x_motor.get_position()) < 1 * q.deg



"""
Photo diode device base class
"""
from concert.quantities import q
from concert.base import Quantity
from concert.devices.base import Device
from concert.base import AccessorNotImplementedError


class PhotoDiode(Device):

    """
    Impementation of photo diode with V output signal
    """

    intensity = Quantity(q.V)

    def __init__(self):
        super(PhotoDiode, self).__init__()

    async def _get_intensity(self):
        raise AccessorNotImplementedError



"""
Measures operate on some data and provide information about how \"good\"
the data is.
"""
import logging
import numpy as np
from concert.quantities import q


LOG = logging.getLogger(__name__)


class DummyGradientMeasure(object):

    """Gradient measure that returns a quadratic fall-off of *parameter* from
    *max_position*."""

    def __init__(self, parameter, max_position, negative=False):
        self.max_position = max_position
        self.negative = negative
        self._max_gradient = 1e4
        self._param = parameter

    async def __call__(self):
        value = await self._param.get()
        position = value.to(self.max_position.units).magnitude
        result = self._max_gradient - (position - self.max_position.magnitude) ** 2

        return -result if self.negative else result


class Area(object):

    """
    Area measures how much of the whole image is covered by an object.
    The return value is the fraction between object pixels and the whole
    area, :math:`n = object\_pixels / (width \cdot height)`.

    Parameters are:

    .. py:attribute:: func

        a callable *func(radio, \*args, \*\*kwargs)*, where
        *radio* is a radiograph to be analyzed, *args* and *kwargs*
        are additional positional and keyword arguments passed to *func*
        which returns the number of pixels covered by
        an object

    .. py:attribute:: args

        *func* arguments tuple

    .. py:attribute:: kwargs

        *func* keyword arguments dictionary

    """

    def __init__(self, func, args=None, kwargs=None):
        self.func = func
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, radio):
        object_pixels = self.func(radio, *self.args, **self.kwargs)

        return object_pixels / (radio.shape[0] * radio.shape[1])


class SimpleArea(Area):

    """
    An area measure with object segmentation by a threshold. If two
    flat fields *flat_1* and *flat_2* are given, then the minimum
    of their subtraction :math:`min(flat\_1, flat\_2)` approximates
    the maximum grey value considered a background. The values
    from the :math:`radiograph - flat` below the minimum are
    considered an object.
    """

    def __init__(self, flat_1, flat_2):
        super(SimpleArea, self).__init__(self.object_pixels)
        self._flat_1 = flat_1
        self._flat_2 = flat_2
        self.flat_avg = None
        self.threshold = None
        self._setup()

    def object_pixels(self, radio):
        """Get object pixels from a radiograph *radio*."""
        radio = radio - self.flat_avg

        return len(np.where(radio < self.threshold)[0])

    @property
    def flats(self):
        """flat_1 used by measure"""
        return [self._flat_1, self._flat_2]

    @flats.setter
    def flats(self, flats):
        """
        Set new flat fields flat_1 and flat_2 from the tuple *flats*.
        The threshold and average flat are recomputed.
        """
        if flats[0].shape != flats[1].shape:
            raise ValueError("Flats must have the same shape")
        self._flat_1 = flats[0]
        self._flat_2 = flats[1]
        self._setup()

    def _setup(self):
        """Setup the threshold and average flat field."""
        # Threshold
        self.threshold = np.min(self._flat_1 - self._flat_2)
        # Average the two flats to suppress the noise a little
        self.flat_avg = (self._flat_1 + self._flat_2) / 2


def rotation_axis(tips):
    """
    Determine a 3D circle normal inclination angles towards the
    :math:`y`-axis (0,1,0) and the center of the circle.

    **Algorithm**

    We exploit the fact that the projected circle is an ellipse. Thus,
    ellipse's major axis is the radius of rotation. If we set the beam
    direction to be the :math:`z`-axis :math:`(0,0,1)`, we can determine
    the :math:`xy` angle :math:`(\phi)` by the major axis angle
    of the ellipse. Furthermore, based on the ratio between minor
    and major axis, we can determine the projected circle's
    inclination from its normal, yielding the :math:`zy` angle
    :math:`(\psi)` of the circle. One then needs to rotate the the sample
    :math:`(-\phi,\pm \psi)`. The :math:`\psi` angle ambiguity comes from
    the fact that we are working with projected data.

    There are three possible situations with the ellipse fitting algorithm:

    * The sample does not move in the image sequence, it is not off-centered
        enough.
    * The ellipse degenerates to a line, which means the circle is parallel
        to the beam, thus we only need to rotate by :math:`-\phi`.
    * The projections lead to a non-degenerate ellipse. In this case we need to
        rotate around both angles.

    **Input**

    An image sequence of at least 5 images with different sample positions
    around the axis of rotation. It is highly recommended to use more images
    in order to obtain more precise results from noisy data.

    **Steps**

    The measuring procedure is done in the following steps:

    #. Sample extraction.
    #. Sample tip determination (largest Euclidean distance from the sample
        entering to the image and its boundaries).
    #. Ellipse fitting to the set of tip points.
    #. Determine :math:`\phi` and :math:`\psi` from the projected circle.

    **Assumptions**

    The procedure assumes a high-absorbing sample with conic shape. Thanks to
    such sample the segmentation procedure can be simplified to a thresholding
    technique.

    **Output**

    The measure returns a tuple (:math:`\phi`, :math:`\psi`, center).
    """
    if len(tips) < 5:
        raise ValueError("At least 5 coordinate pairs are needed")

    y_ind, x_ind = list(zip(*tips))
    if max(x_ind) - min(x_ind) < 10:
        raise ValueError("Sample off-centering too small, enlarge rotation radius.")

    a_matrix = np.empty((len(x_ind), 6), dtype=np.float)
    for i in range(len(x_ind)):
        a_matrix[i] = np.array([x_ind[i] ** 2, x_ind[i] * y_ind[i],
                                y_ind[i] ** 2, x_ind[i], y_ind[i], 1])

    v_mat = np.linalg.svd(a_matrix)[2]
    params = v_mat.T[:, -1]
    a_33 = np.array([[params[0], params[1] / 2], [params[1] / 2, params[2]]])
    x_ind = np.array(x_ind)
    y_ind = np.array(y_ind)
    d_x = x_ind.max() - x_ind.min()
    usv = np.linalg.svd(a_33)
    s_vec = usv[1]
    v_mat = usv[2]
    det = np.linalg.det(a_33)
    LOG.debug('Conic determinant: %g', det)

    if det <= 0:
        # Not an ellipse
        d_y = float(y_ind.max() - y_ind.min())
        angle = np.arctan(d_y / d_x) * q.rad
        if x_ind[y_ind.argmax()] <= x_ind[y_ind.argmin()]:
            # More than pi/4.
            angle = -angle
        phi, psi = (angle, 0.0 * q.rad)
        center = np.mean(tips, axis=0)
    else:
        x_pos = ((params[1] * params[4] - 2 * params[3] * params[2])
                 / (4 * params[0] * params[2] - params[1] ** 2))
        y_pos = ((params[3] * params[1] - 2 * params[0] * params[4])
                 / (4 * params[0] * params[2] - params[1] ** 2))
        center = np.array((y_pos, x_pos))
        # Determine rotation direction needed for the pitch angle
        v_0 = np.array(tips[0]) - center
        v_1 = np.array(tips[1]) - center
        pitch_d_angle = np.arctan2(np.cross(v_0, v_1), np.dot(v_0, v_1))
        sgn = int(np.sign(pitch_d_angle))
        phi, psi = (np.arctan(v_mat[1][1] / v_mat[1][0]) * q.rad,
                    sgn * np.arcsin(np.sqrt(s_vec[1]) / np.sqrt(s_vec[0])) * q.rad)

    phi = phi.to(q.deg)
    psi = psi.to(q.deg)

    LOG.debug('Found z angle: %s, x angle: %s, center: %s', phi, psi, center)

    return (phi, psi, center)



"""Monochromator Dummy"""
from concert.quantities import q
from concert.devices.monochromators import base


class Monochromator(base.Monochromator):

    """Monochromator class implementation."""

    def __init__(self):
        super(Monochromator, self).__init__()
        self._energy = 100 * q.keV

    async def _get_energy_real(self):
        return self._energy

    async def _set_energy_real(self, energy):
        self._energy = energy
